/* $Id: Cerdisp.c 1086 2006-06-06 10:45:30Z d $ */
/* David Leonard, 2001. Public domain. */

#include <stdio.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>

#include <netdb.h>
#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Shell.h>
#include <X11/Xos.h>
#include <X11/Xmd.h>
#include <X11/keysym.h>

#include "CerdispP.h"
#include "CerdispS.h"
#include "lru.h"

/* Define socklen_t for platforms that do not have it */
#ifdef __APPLE__
# include <AvailabilityMacros.h>
# if MAC_OS_X_VERSION_MAX_ALLOWED <= MAC_OS_X_VERSION_10_3
typedef int socklen_t;
# endif
#endif

/* Make sure we have getaddrinfo() */
#ifndef AI_PASSIVE
#include "getaddrinfo-fake.c"
#endif

/* Allow gcc to optomise dead (non-returning) functions */
#ifdef __GNUC__
# define DEAD	__attribute__((noreturn))
#else
# define DEAD	/* */
#endif

/* Input state (for the state and substate fields) */
#define STATE_PREAMBLELEN	1	/* reading preamble length word */
#define STATE_PREAMBLE		2	/* reading preamble */
#define STATE_PALETTEHDR	3	/* reading palette header */
#define STATE_PALETTE		4	/* reading palette */
#define STATE_FRAMEHDR		5	/* reading frame header */
#define STATE_FRAMEDATA		6	/* reading frame data */

/* Little endian to host conversion macros */
#undef letoh32
#undef letoh16
#undef htole32
#undef htole16
#if X_BYTE_ORDER == X_LITTLE_ENDIAN
# define letoh16(x)	((CARD16)(x))
# define letoh32(x)	((CARD32)(x))
# define htole16(x)	((CARD16)(x))
# define htole32(x)	((CARD32)(x))
#else /* X_BIG_ENDIAN */
static CARD16 myswap16(CARD16 x) {
	union { CARD16 i; CARD8 c[2]; } u1, u2;
	u1.i = x;
	u2.c[1] = u1.c[0];
	u2.c[0] = u1.c[1];
	return u2.i;
}
static CARD32 myswap32(CARD32 x) {
	union { CARD32 i; CARD8 c[4]; } u1, u2;
	u1.i = x;
	u2.c[3] = u1.c[0];
	u2.c[2] = u1.c[1];
	u2.c[1] = u1.c[2];
	u2.c[0] = u1.c[3];
	return u2.i;
}
# define letoh16(x)	myswap16(x)
# define letoh32(x)	myswap32(x)
# define htole16(x)	myswap16(x)
# define htole32(x)	myswap32(x)
#endif /* X_BIG_ENDIAN */

#define InvalidInputID	(-1)

/* Ugh! this is needed to add socket input handling */
extern XtAppContext ac;

/*
 * Socket data handlers
 */
static int handle_preamblelen(CerdispWidget, char *);
static int handle_preamble(CerdispWidget, char *);
static int handle_palettehdr(CerdispWidget, char *);
static int handle_palette(CerdispWidget, char *);
static int handle_framehdr(CerdispWidget, char *);
static int handle_framedata(CerdispWidget, char *, int);

/*
 * Class Methods
 */
static void CerdispInitialize(Widget, Widget, ArgList, Cardinal*);
static void CerdispExpose(Widget, XEvent *, Region);
static void CerdispRealize(Widget, Mask *, XSetWindowAttributes *);
static Boolean CerdispSetValues(Widget, Widget, Widget, ArgList, Cardinal *);
static void CerdispDestroy(Widget);

/*
 * Misc
 */
static void clearwindow(CerdispWidget);
static void settitle(CerdispWidget);
static void writebufto(int, void *, size_t);
static void proto_fail(Widget, const char *);
static void proto_failx(Widget, const char *, ...);
static void fatal_error(const char *fmt, ...) DEAD;
static void proxyinput(XtPointer, int *, XtInputId *);
static void configure_image(CerdispWidget tw);
static unsigned long palette_lookup(CerdispWidget, unsigned int);

/*
 * Actions
 */
static void InputAction(Widget, XEvent*, String*, Cardinal*);
static void CloseAction(Widget, XEvent*, String*, Cardinal*);
static void ScaleAction(Widget, XEvent*, String*, Cardinal*);

/*
 * Constants
 */
static const int one = 1;

/*
 * Initialization
 */
#define offset(field) XtOffsetOf(CerdispRec, cerdisp.field)
static XtResource resources[] = {
/*{ name, class, type,
    size, offset,
    default_type, default_addr },*/
  { XtNproxy, XtCProxy, XtRString,
    sizeof(char *), offset(proxy),
    XtRImmediate, (XtPointer)NULL },
  { XtNlistenPort, XtCListenPort, XtRString,
    sizeof(char *), offset(listen_port),
    XtRImmediate, (XtPointer)"987" },
  { XtNproxyPort, XtCProxyPort, XtRString,
    sizeof(char *), offset(proxy_port),
    XtRImmediate, (XtPointer)"987" },
  { XtNprintVersion, XtCPrintVersion, XtRBoolean,
    sizeof(Boolean), offset(print_version),
    XtRImmediate, (XtPointer)False },
  { XtNverbose, XtCVerbose, XtRBoolean,
    sizeof(Boolean), offset(verbose),
    XtRImmediate, (XtPointer)False },
  { XtNscale, XtCScale, XtRInt,
    sizeof(int), offset(scale),
    XtRInt, (XtPointer)&one },
};
#undef offset

/*
 * Action table for the widget
 */
static XtActionsRec actions[] =
{
    /*{name,	procedure},*/
    {"input",	InputAction},
    {"close",	CloseAction},
    {"scale",	ScaleAction},
};

/*
 * Event translation table for the widget
 */
static char translations[] =
"!Shift<Key>KP_Add:"		"scale(+1)\n"
"!Shift<Key>KP_Subtract:"	"scale(-1)\n"
"<KeyDown>:"			"input()\n"
"<KeyUp>:"			"input()\n"
"<Btn1Down>:"			"input()\n"
"<Btn1Motion>:"			"input()\n"
"<Btn1Up>:"			"input()\n"
;

/*
 * Class record (methods and properties)
 */
CerdispClassRec cerdispClassRec = {
  /* core */
  {
    (WidgetClass)&widgetClassRec,	/* superclass */
    "Cerdisp",				/* class_name */
    sizeof(CerdispRec),			/* widget_size */
    NULL,				/* class_initialize */
    NULL,				/* class_part_initialize */
    False,				/* class_inited */
    CerdispInitialize,			/* initialize */
    NULL,				/* initialize_hook */
    CerdispRealize,			/* realize */
    actions,				/* actions */
    XtNumber(actions),			/* num_actions */
    resources,				/* resources */
    XtNumber(resources),		/* num_resources */
    NULLQUARK,				/* xrm_class */
    True,				/* compress_motion */
    True,				/* compress_exposure */
    True,				/* compress_enterleave */
    False,				/* visible_interest */
    CerdispDestroy,			/* destroy */
    XtInheritResize,			/* resize */
    CerdispExpose,			/* expose */
    CerdispSetValues,			/* set_values */
    NULL,				/* set_values_hook */
    XtInheritSetValuesAlmost,		/* set_values_almost */
    NULL,				/* get_values_hook */
    XtInheritAcceptFocus,		/* accept_focus */
    XtVersion,				/* version */
    NULL,				/* callback_private */
    translations,			/* tm_table */
    XtInheritQueryGeometry,		/* query_geometry */
    XtInheritDisplayAccelerator,	/* display_accelerator */
    NULL,				/* extension */
  },
  /* cerdisp */
  {
    NULL,				/* extension */
  }
};

WidgetClass cerdispWidgetClass = (WidgetClass)&cerdispClassRec;

/*
 * Implementation
 */

/* Update the window title to reflect the connection */
static void
settitle(CerdispWidget tw)
{
    Arg arglist[2];
    const char *icontitle;
    const char *title;

    if (tw->cerdisp.connected && tw->cerdisp.peername[0] != '\0') {
	title = tw->cerdisp.peername;
	icontitle = tw->cerdisp.peername;
    } else if (tw->cerdisp.connected) {
	title = "Xcerdisp: connecting...";
	icontitle = "xcerdisp";
    } else {
	title = "Xcerdisp: not connected";
	icontitle = "xcerdisp";
    }
    XtSetArg(arglist[0], XtNtitle, title);
    XtSetArg(arglist[1], XtNiconName, icontitle);
    XtSetValues(XtParent((Widget)tw), arglist, 2);
}

/* Fill the window with the default colour */
static void
clearwindow(CerdispWidget tw)
{
    Widget w = (Widget)tw;
    unsigned int width, height;
    unsigned int ignore;
    Window wignore;

    XGetGeometry(XtDisplay(w), XtWindow(w), &wignore, &ignore, &ignore, 
	&width,  &height, &ignore, &ignore);
    XFillRectangle(XtDisplay(w), XtWindow(w), tw->cerdisp.gc, 0, 0, 
	width, height);
}

/* Write a buffer to a socket, retrying when partial writes occur */
static void
writebufto(int fd, void *data, size_t len)
{
    int slen;
    char *buf = data;

    while (len > 0) {
	slen = write(fd, buf, len);
	if (slen == -1)
	    fatal_error("Could not send data: %s", strerror(errno));
	buf += slen;
	len -= slen;
    }
}

/* Non-fatal system error - closes connection */
static void
proto_fail(Widget w, const char *msg)
{
    return proto_failx(w, "%s: %s", msg, strerror(errno));
}

/* Protocol (non-system) error - closes connection */
static void
proto_failx(Widget w, const char *fmt, ...)
{
    va_list ap;
    char buf[1024];

    va_start(ap, fmt);
    vsnprintf(buf, sizeof buf, fmt, ap);
    va_end(ap);
    XtAppWarning(ac, buf);
    CloseAction(w, NULL, NULL, NULL);
}

static void
fatal_error(const char *fmt, ...)
{
    va_list ap;
    char buf[1024];

    va_start(ap, fmt);
    vsnprintf(buf, sizeof buf, fmt, ap);
    va_end(ap);
    XtAppError(ac, buf);
    /* NOTREACHED */
    exit(1);
}

/* Called when data is received on our proxy port */
static void
proxyinput(XtPointer closure, int *source, XtInputId *id)
{
    Widget w = (Widget)closure;
    CerdispWidget tw = (CerdispWidget)w;
    unsigned char buf[1024];
    int len;

    len = read(*source, buf, sizeof buf);
    if (len == -1) {
	proto_fail(w, "proxyinput: read");
	return;
    }
    if (len == 0) {
	if (tw->cerdisp.verbose)
		fprintf(stderr, "proxyinput: connection closed\n");
        CloseAction(w, NULL, NULL, NULL);
	return;
    }

    /* Copy the data from the windows server to the handheld */
    /* XXX - assumes local input is quiescent or data interleaves */
    writebufto(tw->cerdisp.clientsock, buf, len);
}

/* Allocate and clear the in-memory XImage and resize the widget */
static void
configure_image(CerdispWidget tw)
{
    Widget w = (Widget)tw;
    int width, height, scale;
    Visual *visual;
    XImage *oldimage;
    Arg arglist[4];
    int x, y;

    /* Make sure the scale is sane */
    scale = tw->cerdisp.scale;
    if (scale <= 0) {
        XtAppWarning(ac, "Invalid Cerdisp.scale value, changing to 1");
        scale = tw->cerdisp.scale = 1;
    }

    /* Change the size of this widget */
    width = tw->cerdisp.frame_width * scale;
    height = tw->cerdisp.frame_height * scale;
    XtSetArg(arglist[0], XtNwidth, width);
    XtSetArg(arglist[1], XtNheight, height);
    XtSetValues(w, arglist, 2);

    /* Constrain the shell to minimum width and height */
    XtSetArg(arglist[0], XtNminWidth, width);
    XtSetArg(arglist[1], XtNminHeight, height);
    XtSetArg(arglist[2], XtNmaxWidth, width);
    XtSetArg(arglist[3], XtNmaxHeight, height);
    XtSetValues(XtParent(w), arglist, 4);

    /* Get the parent's visual */
    visual = NULL;
    XtSetArg(arglist[0], XtNvisual, &visual);
    XtGetValues(XtParent(w), arglist, 1);
    if (visual == NULL)
	visual = DefaultVisualOfScreen(XtScreen(w));

    /* Save any old image */
    oldimage = tw->cerdisp.image;

    /* Allocate an XImage */
    tw->cerdisp.image = XCreateImage(
	XtDisplay(w),		/* display */
	visual,			/* visual */
	tw->core.depth,		/* depth */
	ZPixmap,		/* format */
	0,			/* offset */
	NULL,			/* data */
	width,			/* width */
	height,			/* height */
	8,			/* bitmap_pad */
	0			/* bytes_per_line */
    );
    if (tw->cerdisp.image == NULL)
	fatal_error("XCreateImage failed on visual 0x%x",
	    XVisualIDFromVisual(visual));

    /* Allocate storage for the image, and clear to zero */
    tw->cerdisp.image->data = XtCalloc(height,
	tw->cerdisp.image->bytes_per_line);

    if (oldimage) {
        /* Copy the old image into the new image */
	for (y = 0; y < height; y++)
	  for (x = 0; x < width; x++)
	    XPutPixel(tw->cerdisp.image, x, y, XGetPixel(oldimage, 
		 oldimage->width * x / width, oldimage->height * y / height));
	XDestroyImage(oldimage);
    }
}

/* Handle some input data from the client */
static void
CerdispClientInput(XtPointer closure, int *source, XtInputId *id)
{
    Widget w = (Widget)closure;
    CerdispWidget tw = (CerdispWidget)w;
    CerdispPart *cp = &tw->cerdisp;
    int len;

    /*
     * We read partial data into a buffer until we receive
     * the expected number of bytes.
     */
    if (cp->expected > 0) {
	if (cp->bufsz < cp->bufpos + cp->expected) {
		cp->bufsz = cp->expected;
		if (cp->buf == NULL)
			cp->buf = XtMalloc(cp->bufsz);
		else
			cp->buf = XtRealloc(cp->buf, cp->bufsz);
	}
	len = read(*source, cp->buf + cp->bufpos, cp->expected);

	/* Stop immediately if the connection closed. Not an error. */
	if (len == 0) {
		CloseAction(w, NULL, NULL, NULL);
		return;
	}

	/* Read errors shut us down */
	if (len == -1) {
		proto_fail(w, "CerdispClientInput: read");
		/* Note: CloseAction() has been called */
		return;
	}

	/* Copy all data through to any proxy */
	if (cp->proxysock != -1)
		writebufto(cp->proxysock, cp->buf + cp->bufpos, len);

	cp->expected -= len;
	cp->bufpos += len;

	/* Return if there is more data expected */
	if (cp->state != STATE_FRAMEDATA && cp->expected > 0)
		return;
    }

    /* 
     * Finished reading data in a state. The handler should
     * advance the state to what is expected next.
     */
    switch (cp->state) {
    case STATE_PREAMBLELEN:
	len = handle_preamblelen(tw, cp->buf);
	break;
    case STATE_PREAMBLE:
	len = handle_preamble(tw, cp->buf);
	break;
    case STATE_PALETTEHDR:
	len = handle_palettehdr(tw, cp->buf);
	break;
    case STATE_PALETTE:
	len = handle_palette(tw, cp->buf);
	break;
    case STATE_FRAMEHDR:
	len = handle_framehdr(tw, cp->buf);
	break;
    case STATE_FRAMEDATA:
	len = handle_framedata(tw, cp->buf, cp->bufpos);
	break;
    default:
	fatal_error("CerdispClientInput: bad state %d", cp->state);
    }

    /* Sanity check */
    if (len > cp->bufpos)
	fatal_error("CerdispClientInput: handler overconsumed");
    else if (len > 0 && len < cp->bufpos)
	    memmove(cp->buf, cp->buf + len, cp->bufpos - len);
    cp->bufpos -= len;
}

/* Receive the initial preamble length */
static int
handle_preamblelen(CerdispWidget tw, char *data)
{
    CARD32 *lenp = (CARD32 *)data;
    CARD32 len;
    Widget w = (Widget)tw;

    len = letoh32(*lenp);

    if (tw->cerdisp.verbose)
	fprintf(stderr, "preamblelen = %ld\n", len);

    if (len != sizeof (struct preamble) + sizeof len) {
	proto_failx(w, "handle_preamblelen: bad preamble size != %d",
	    sizeof (struct preamble) + sizeof len);
	return 0;
    }

    tw->cerdisp.state = STATE_PREAMBLE;
    tw->cerdisp.expected = sizeof (struct preamble);

    return sizeof *lenp;
}

/* Receive the preamble structure and initialise things */
static int
handle_preamble(CerdispWidget tw, char *data)
{
    Widget w = (Widget)tw;
    struct preamble *p = (struct preamble *)data;

    if (tw->cerdisp.verbose) {
	fprintf(stderr,"preamble:\n");
	fprintf(stderr," magic	 %08lx\n", letoh32(p->magic));
	fprintf(stderr," width	 %8ld\n", letoh32(p->width));
	fprintf(stderr," height	 %8ld\n", letoh32(p->height));
	fprintf(stderr," depth	 %8ld\n", letoh32(p->depth));
	fprintf(stderr," unk1	 %08lx\n", letoh32(p->unk1));
	fprintf(stderr," unk2	 %08lx\n", letoh32(p->unk2));
	fprintf(stderr," platf	 %.128s\n", p->platform);
	fprintf(stderr," devtyp	 %.128s\n", p->devtype);
	fprintf(stderr," name	 %.16s\n", p->name);
	fprintf(stderr," unk3	 %08lx\n", letoh32(p->unk3));
    }

    tw->cerdisp.frame_width = letoh32(p->width);
    tw->cerdisp.frame_height = letoh32(p->height);

    /* Keep a record of the depth of the source */
    tw->cerdisp.depth = letoh32(p->depth);

    /* Allocate an image */
    configure_image(tw);

    /* Generate an expose event to draw the (blank) image */
    XClearWindow(XtDisplay(w), XtWindow(w));

    /* Change the window title */
    snprintf(tw->cerdisp.peername, sizeof tw->cerdisp.peername, p->name);
    settitle(tw);

    /* Depending on the preamble, we next expect either a palette or a frame */
    if (letoh32(p->depth) == 16) {
	tw->cerdisp.state = STATE_FRAMEHDR;
	tw->cerdisp.expected = sizeof (struct framehdr);
    } else {
	tw->cerdisp.state = STATE_PALETTEHDR;
	tw->cerdisp.expected = sizeof (struct palettehdr);
    }

    return sizeof *p;
}

/* Receive the palette header */
static int
handle_palettehdr(CerdispWidget tw, char *data)
{
    Widget w = (Widget)tw;
    struct palettehdr *ph = (struct palettehdr *)data;
    int datalen = letoh32(ph->datalen);
    int len = 1 << tw->cerdisp.depth;

    if (tw->cerdisp.verbose) {
	fprintf(stderr,"palettehdr:\n");
	fprintf(stderr," type	 %08lx\n", letoh32(ph->type));
	fprintf(stderr," datalen %08lx\n", letoh32(ph->datalen));
    }

    if (letoh32(ph->type) != 1 || 
	datalen != len * sizeof (struct palette))
    {
	proto_failx(w, "handle_palettehdr: bad header");
	return 0;
    }

    tw->cerdisp.state = STATE_PALETTE;
    tw->cerdisp.expected = datalen;

    return sizeof *ph;
}

/* Receive the palette data */
static int
handle_palette(CerdispWidget tw, char *data)
{
    struct palette *palette = (struct palette *)data;
    int ncols = 1 << tw->cerdisp.depth;
    int i;

    tw->cerdisp.palette = (struct palette_entry *)XtCalloc(ncols,
	sizeof (struct palette_entry));

    /* Convert the received palette into our own palette structure */
    for (i = 0; i < ncols; i++) {
	/* I guess only 5 bits are valid - this looks much better */
	tw->cerdisp.palette[i].red = palette[i].red >> 3;
	tw->cerdisp.palette[i].green = palette[i].green >> 3;
	tw->cerdisp.palette[i].blue = palette[i].blue >> 3;
    }

    tw->cerdisp.state = STATE_FRAMEHDR;
    tw->cerdisp.expected = sizeof (struct framehdr);

    return ncols * sizeof *palette;
}

/* Receive a frame header */
static int
handle_framehdr(CerdispWidget tw, char *data)
{
    Widget w = (Widget)tw;
    struct framehdr *fh = (struct framehdr *)data;

    if (tw->cerdisp.verbose) {
	fprintf(stderr,"framehdr:\n");
	fprintf(stderr," unk0	 %08lx\n", letoh32(fh->unk0));
	fprintf(stderr," datalen %08lx\n", letoh32(fh->datalen));
	fprintf(stderr," unk1	 %08lx\n", letoh32(fh->unk1));
	fprintf(stderr," unk2	 %08lx\n", letoh32(fh->unk2));
	fprintf(stderr," width	 %08lx\n", letoh32(fh->width));
	fprintf(stderr," height	 %08lx\n", letoh32(fh->height));
    }

    /* 
     * XXX - I think unk1, unk2 are supposed to be offsets
     * into the image, and width/height are the size of the
     * rectangle of data, so as to allow partial frames to be
     * sent. But I've never seen it happen! We check, nonetheless.
     */
    if (letoh32(fh->width) != tw->cerdisp.frame_width
	|| letoh32(fh->height) != tw->cerdisp.frame_height)
    {
	    proto_failx(w, "handle_framehdr: bad frame dimension");
	    return 0;
    }

    tw->cerdisp.state = STATE_FRAMEDATA;
    tw->cerdisp.expected = letoh32(fh->datalen) * sizeof (CARD16);

    tw->cerdisp.framepos = 0;

    return sizeof *fh;
}

/* Receive (partial) frame data */
static int
handle_framedata(CerdispWidget tw, char *data, int size)
{
    Widget w = (Widget)tw;
    int i, j, k, dx, dy;
    int x, y, starty, endy;
    int len;
    int scale = tw->cerdisp.scale;
    int maxpos = tw->cerdisp.frame_width * tw->cerdisp.frame_height;
    struct rle *rle = (struct rle *)data;
    struct rle current;
    unsigned int repeat;
    unsigned long pixel[4];
    int npixels;

    /* Sanity check; compiler should optomise out */
    if (sizeof *rle != 2 * sizeof (CARD16))
	fatal_error("sizeof *rle = %d", sizeof *rle);

    if (scale <= 0)
	scale = 1;

    /* Remember which row we start at */
    starty = (tw->cerdisp.framepos / tw->cerdisp.frame_width) * scale;

    /* Decode the RLE frame data */
    for (i = 0; i < size / sizeof *rle; i++) {

	current.u.word = letoh16(rle[i].u.word);
	repeat = letoh16(rle[i].repeat);

	/* Extract the set of pixels from the RLE entry */
	if (tw->cerdisp.depth == 4) {
	    npixels = 4;
	    pixel[0] = palette_lookup(tw, current.u.col4.p0);
	    pixel[1] = palette_lookup(tw, current.u.col4.p1);
	    pixel[2] = palette_lookup(tw, current.u.col4.p2);
	    pixel[3] = palette_lookup(tw, current.u.col4.p3);
	} else {
	    npixels = 1;
	    pixel[0] = lrupixel(tw->cerdisp.lru, current.u.col.red,
		current.u.col.green, current.u.col.blue);
	}

	for (j = 0; j < repeat; j++) {
	    for (k = 0; k < npixels; k++) {
		unsigned long colour = pixel[k];

		if (tw->cerdisp.framepos >= maxpos) {
		    XtAppWarning(ac, "handle_framedata(): too much data");
		    break;
		}

		x = (tw->cerdisp.framepos % tw->cerdisp.frame_width) * scale;
		y = (tw->cerdisp.framepos / tw->cerdisp.frame_width) * scale;

		for (dx = 0; dx < scale; dx++)
		    for (dy = 0; dy < scale; dy++)
			XPutPixel(tw->cerdisp.image, x+dx, y+dy, colour);

		tw->cerdisp.framepos++;
	    }
	}
    }

    /* Figure out how much data we consumed */
    len = i * sizeof *rle;

    /* Remember which row we ended at */
    endy = (tw->cerdisp.framepos / tw->cerdisp.frame_width) * scale;

    /* Update the displayed image - XXX should synthesize expose event? */
    XPutImage(XtDisplay(w), XtWindow(w), tw->cerdisp.gc, tw->cerdisp.image,
	0, starty, 0, starty, tw->cerdisp.image->width, endy-starty + 1);

    /* Are we at the end of the frame? */
    if (tw->cerdisp.framepos >= maxpos) {
	lrufreeold(tw->cerdisp.lru);	/* Free old colours */
	tw->cerdisp.state = STATE_FRAMEHDR;
	tw->cerdisp.expected = sizeof (struct framehdr);
    }

    return len;
}

/* Handle a new connection on the server socket */
static void
CerdispAccept(XtPointer closure, int *source, XtInputId *id)
{
    Widget w = (Widget)closure;
    CerdispWidget tw = (CerdispWidget)w;
    char addr[1024];
    socklen_t len;		/* (edit Imakefile if socklen_t bothers you) */
    int s;
    int i;
    char hbuf[NI_MAXHOST], sbuf[NI_MAXSERV];

    len = sizeof addr;
    s = accept(*source, (struct sockaddr *)addr, &len);
    if (s == -1) {
	perror("accept");
	return;
    }

    if (getnameinfo((struct sockaddr *)addr, len, 
	hbuf, sizeof hbuf, sbuf, sizeof sbuf,
	NI_NUMERICHOST | NI_NUMERICSERV))
	    fprintf(stderr, "accept: getnameinfo failed\n");
    else if (tw->cerdisp.verbose)
	    fprintf(stderr, "connection from %s port %s\n", hbuf, sbuf);

    if (tw->cerdisp.proxy && *tw->cerdisp.proxy) { 
	/*
	 * Open up a connection to a windows host running
	 * a real version of the cerdisp, or another xcerdisp
	 * server.
	 */
	struct addrinfo hints, *res, *res0;
	int error;
	int t;
	const char *cause = NULL;

	memset(&hints, 0, sizeof hints);
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	error = getaddrinfo(tw->cerdisp.proxy, tw->cerdisp.proxy_port, &hints, &res0);
	if (error) {
	    fatal_error("getaddrinfo: %s:%s: %s\n", tw->cerdisp.proxy,
		tw->cerdisp.proxy_port, gai_strerror(error));
	}
	t = -1;
	for (res = res0; res; res = res->ai_next) {
	    t = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	    if (t < 0) {
		cause = "socket";
		continue;
	    }

	    if (getnameinfo(res->ai_addr, res->ai_addrlen,
		hbuf, sizeof hbuf, sbuf, sizeof sbuf,
		NI_NUMERICHOST | NI_NUMERICSERV))
		    fprintf(stderr, "getnameinfo failed\n");
	    else if (tw->cerdisp.verbose)
		    fprintf(stderr, "connecting to proxy at %s port %s\n", hbuf, sbuf);

	    if (connect(t, res->ai_addr, res->ai_addrlen) < 0) {
		cause = "connect";
		close(t);
		t = -1;
		continue;
	    }
	    break;
	}
	freeaddrinfo(res0);

	tw->cerdisp.proxysock = t;
	if (t == -1) {
	    perror(cause);
	    XtAppWarning(ac, "Could not connect to proxy");
	} else {
	    if (tw->cerdisp.verbose)
		    fprintf(stderr, "connected to proxy\n");
	    tw->cerdisp.proxyinput = XtAppAddInput(ac, t, 
		(XtPointer)XtInputReadMask, proxyinput, (XtPointer)tw);
	}
    } else
	tw->cerdisp.proxysock = -1;

    /* Stop listening on the server socket */
    for (i = 0; i < MAXSOCK; i++)
	if (tw->cerdisp.serverinput[i] != InvalidInputID) {
            XtRemoveInput(tw->cerdisp.serverinput[i]);
            tw->cerdisp.serverinput[i] = InvalidInputID;
        }

    /* Start listening on the client socket */ 
    tw->cerdisp.clientinput = XtAppAddInput(ac, s, (XtPointer)XtInputReadMask, 
	CerdispClientInput, (XtPointer)tw);

    /* Indicate that we are connected */
    tw->cerdisp.clientsock = s;
    tw->cerdisp.connected = 1;
    tw->cerdisp.peername[0] = '\0';

    /* Wait for the preamble length */
    tw->cerdisp.state = STATE_PREAMBLELEN;
    tw->cerdisp.expected = sizeof (CARD32);

    /* Set the title to 'connecting...' */
    settitle(tw);
}

/* Initialise widget instance */
/*ARGSUSED*/
static void
CerdispInitialize(Widget request, Widget w, ArgList args, Cardinal *num_args)
{
    CerdispWidget tw = (CerdispWidget)w;
    struct addrinfo hints, *res, *res0;
    int error;
    int nsock;
    const char *cause = NULL;
    int s, i;

    /* Mention our version number for debugging */
    if (tw->cerdisp.print_version) {
	fprintf(stderr, "Xcerdisp version %s\n", CERDISP_VERSION);
    }

    /* Invalidate all socket slots */
    for (i = 0; i < MAXSOCK; i++) {
        tw->cerdisp.serversock[i] = -1;
	tw->cerdisp.serverinput[i] = InvalidInputID;
    }

    /* Open sockets to listen on */
    memset(&hints, 0, sizeof hints);
    hints.ai_family = PF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_PASSIVE;
    error = getaddrinfo(NULL, tw->cerdisp.listen_port, &hints, &res0);
    if (error)
	fatal_error("Cannot create server endpoint: %s", gai_strerror(error));
    nsock = 0;
    for (res = res0; res && nsock < MAXSOCK; res = res->ai_next) {
	s = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	if (s < 0) {
	    cause = "socket";
	    continue;
	}

#ifdef SO_REUSEADDR
	/*
	 * Allow local address reuse. Only meaningful when the TCP stack
	 * holds an endpoint in FIN_WAIT_* and the user is impatient.
	 */
	/* XXX Why doesn't C allow constant addressing, i.e. &1 */
	i = 1;
	(void)setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &i, sizeof i);
#endif

	/* Bind and listen */
	if (bind(s, res->ai_addr, res->ai_addrlen) < 0) {
	    cause = "bind";
	    close(s);
	    continue;
	}
	listen(s, 5);

	/* 
	 * Keepalives consist of the kernel sending periodic,
	 * redundant TCP ACKs. This is useful because when the PocketPC
	 * is undocked, WinCE closes all its local endpoints, but
	 * we are left half-open. Keepalives eventually cause a 
	 * timeout+close in this case.
	 */
        i = 1;
	setsockopt(s, SOL_SOCKET, SO_KEEPALIVE, &i, sizeof i);

#if 0
	{
	    /* Some debugging to print our local listening endpoint */
	    char hbuf[NI_MAXHOST], sbuf[NI_MAXSERV];
	    if (getnameinfo(res->ai_addr, res->ai_addrlen, 
		hbuf, sizeof hbuf, sbuf, sizeof sbuf,
		NI_NUMERICHOST | NI_NUMERICSERV))
		    fprintf(stderr, "getnameinfo failed\n");
	    else if (tw->cerdisp.verbose)
		    fprintf(stderr, "bound to %s : %s\n", hbuf, sbuf);
	}
#endif

        tw->cerdisp.serversock[nsock++] = s;
    }
    if (nsock == 0) 
	fatal_error("Cannot create server endpoint: %s: %s", cause,
	    strerror(errno));
    freeaddrinfo(res0);

    /* Reset state to 'unconnected' */
    tw->cerdisp.connected = 0;
    tw->cerdisp.clientsock = -1;
    tw->cerdisp.clientinput = InvalidInputID;

    /* Zero resource pointers */
    tw->cerdisp.image = NULL;
    tw->cerdisp.bufsz = 0;
    tw->cerdisp.buf = NULL;
    tw->cerdisp.palette = NULL;

    /* Allocate colours */
    tw->cerdisp.lru = lrualloc((Widget)tw);

    /* Give our window a title */
    settitle(tw);

    /* Add handlers to accept incoming client connections. */
    for (i = 0; i < MAXSOCK; i++)
	if (tw->cerdisp.serversock[i] != -1)
	    tw->cerdisp.serverinput[i] = XtAppAddInput(ac,
		tw->cerdisp.serversock[i], (XtPointer)XtInputReadMask,
		CerdispAccept, (XtPointer)tw);

    /* Mention our proxy's listen address */
    if (tw->cerdisp.proxy && *tw->cerdisp.proxy) { 
	fprintf(stderr, "proxy at %s:%s\n", tw->cerdisp.proxy,
	    tw->cerdisp.proxy_port);
    }
}

/* Handle an expose event */
static void
CerdispExpose(Widget w, XEvent *e, Region r)
{
    CerdispWidget tw = (CerdispWidget)w;
    XRectangle rect;
    Display *display;

    display = XtDisplay(w);
    XClipBox(r, &rect);
    if (!tw->cerdisp.connected || tw->cerdisp.image == NULL)
	/* fill with fg colour when not connected */
	XFillRectangles(display, XtWindow(w), tw->cerdisp.gc, &rect, 1);
    else
	/* redraw previous pixels */
	XPutImage(display, XtWindow(w), tw->cerdisp.gc, tw->cerdisp.image,
	    rect.x, rect.y, rect.x, rect.y, rect.width, rect.height);
}

/* Realize the widget */
static void
CerdispRealize(Widget w, Mask *valueMask, XSetWindowAttributes *attributes)
{
    CerdispWidget tw = (CerdispWidget)w;

    XtCreateWindow(w, (unsigned int)InputOutput, (Visual *)CopyFromParent,
	*valueMask, attributes);
    tw->cerdisp.gc = XtGetGC(w, 0, NULL);
}

/* Handle resource changes in the widget */
static Boolean
CerdispSetValues(Widget current, Widget request, Widget w,
    ArgList args, Cardinal *num_args)
{
    CerdispWidget tcurrent = (CerdispWidget)current;
    CerdispWidget trequest = (CerdispWidget)request;
    CerdispWidget tw = (CerdispWidget)w;
    Boolean redisplay = False;

    /* Only allow non-zero, positive scales */
    if (trequest->cerdisp.scale < 1) {
        XtAppWarning(ac, "Cerdisp: invalid scale; using 1");
	tw->cerdisp.scale = 1;
    } else
	tw->cerdisp.scale = trequest->cerdisp.scale;

    if (tcurrent->cerdisp.scale != tw->cerdisp.scale &&
        tw->cerdisp.image != NULL)
    {
	configure_image(tw);
	redisplay = True;
    }
    return redisplay;
}

/* Destroy the widget */
static void
CerdispDestroy(Widget w)
{
    CerdispWidget tw = (CerdispWidget)w;
    int i;

    /* shutdown the server socket */
    for (i = 0; i < MAXSOCK; i++) {
        if (tw->cerdisp.serverinput[i] != InvalidInputID) {
		XtRemoveInput(tw->cerdisp.serverinput[i]);
		tw->cerdisp.serverinput[i] = InvalidInputID;
	}
        if (tw->cerdisp.serversock[i] != -1) {
            close(tw->cerdisp.serversock[i]);
            tw->cerdisp.serversock[i] = -1;
	}
    }

    /* Make sure resources are released */
    CloseAction(w, NULL, NULL, NULL);

    /* Deallocate colours */
    lrufree(tw->cerdisp.lru);
    tw->cerdisp.lru = NULL;

    XtReleaseGC(w, tw->cerdisp.gc);
}

/* Handle a key or mouse motion in the window */
static void
InputAction(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    CerdispWidget tw = (CerdispWidget)w;
    struct {
	CARD16 status;
	CARD16 x;
	CARD16 y;
    } pointer;
    struct {
	CARD16 status;
	CARD16 key;
    } keyboard;
    CARD16 key;
    KeySym keysym;

    /* Table of synthetic action commands for the first argument */
    static struct { 
	int op;
	const char *str;
    } cmdtab[] = {
	{ EVT_KEY_UP,	"KeyUp" },
	{ EVT_KEY_DOWN,	"KeyDown" },
	{ EVT_PTR_UP,	"BtnUp" },
	{ EVT_PTR_DOWN,	"BtnDown" },
	{ EVT_PTR_MOVE,	"Motion" }
    };

    if (!tw->cerdisp.connected)
	return;

    /* Action has arguments; we generate a synthetic event */
    if (*num_params > 0) {
	int i, op;

	for (i = 0; i < XtNumber(cmdtab); i++)
	    if (strcmp(params[0], cmdtab[i].str) == 0)
		break;
	if (i >= XtNumber(cmdtab)) {
	    XtAppWarning(ac, "input(): bad argument");
	    return;
	}
	op = cmdtab[i].op;
	switch (op) {
	case EVT_KEY_UP:
	case EVT_KEY_DOWN:
	    if (*num_params != 2) {
		XtAppWarning(ac, "input(Key*,): expected 2 arguments");
		return;
	    }
	    keyboard.status = htole16(op);
	    keyboard.key = htole16(strtoul(params[1], NULL, 0));
	    writebufto(tw->cerdisp.clientsock, &keyboard, sizeof keyboard);
	    break;
	case EVT_PTR_UP:
	case EVT_PTR_DOWN:
	case EVT_PTR_MOVE:
	    if (*num_params != 3) {
		XtAppWarning(ac, "input(Btn*|Motion,): expected 3 arguments");
		return;
	    }
	    pointer.status = htole16(op);
	    pointer.x = htole16(strtoul(params[1], NULL, 0));
	    pointer.y = htole16(strtoul(params[2], NULL, 0));
	    writebufto(tw->cerdisp.clientsock, &pointer, sizeof pointer);
	    break;
	}
	return;
    }

    switch (event->type) {
    case ButtonPress:
    case ButtonRelease:
    case MotionNotify:
	pointer.status = htole16(
		event->type == ButtonPress ?    EVT_PTR_DOWN :
		event->type == ButtonRelease ?  EVT_PTR_UP :
						EVT_PTR_MOVE);
	pointer.x = htole16(event->xbutton.x / tw->cerdisp.scale);
	pointer.y = htole16(event->xbutton.y / tw->cerdisp.scale);
	writebufto(tw->cerdisp.clientsock, &pointer, sizeof pointer);
	break;
    case KeyPress:
    case KeyRelease:
	keysym = XLookupKeysym(&event->xkey, 0);

	/* These could have been done with a translation table... */
	switch (keysym) {
	default:
		if (keysym >= XK_a && keysym <= XK_z)
			key = 0x41 + keysym - XK_a;
		else if (keysym >= XK_0 && keysym <= XK_9)
			key = 0x30 + keysym - XK_0;
		else if (keysym >= XK_KP_0 && keysym <= XK_KP_9)
			key = 0x60 + keysym - XK_KP_1;
		else if (keysym >= XK_F1 && keysym <= XK_F9)
			key = 0x70 + keysym - XK_F1;
		else
			key = 0x00;
		break;
	case XK_KP_Prior:
	case XK_Prior:		key = 0x21; break;
	case XK_KP_Next:
	case XK_Next:		key = 0x22; break;
	case XK_End:
	case XK_KP_End:		key = 0x23; break;
	case XK_Home:
	case XK_KP_Home:	key = 0x24; break;
	case XK_Left:
	case XK_KP_Left:	key = 0x25; break;
	case XK_Up:
	case XK_KP_Up:		key = 0x26; break;
	case XK_Right:
	case XK_KP_Right:	key = 0x27; break;
	case XK_Down:
	case XK_KP_Down:	key = 0x28; break;
	case XK_Insert:
	case XK_KP_Insert:	key = 0x2d; break;
	case XK_Delete:	
	case XK_KP_Delete:	key = 0x2e; break;

	case XK_KP_Begin:	key = 0x0c; break;
	case XK_KP_Multiply:	key = 0x6a; break;
	case XK_KP_Add:		key = 0x6b; break;
	case XK_KP_Subtract:	key = 0x6d; break;
	case XK_KP_Divide:	key = 0x6f; break;

	case XK_KP_Enter:
	case XK_Return:		key = 0x0d; break;

	case XK_Escape:		key = 0x1b; break;
	case XK_BackSpace:	key = 0x08; break;
	case XK_Tab:		key = 0x09; break;

	case XK_Shift_R:
	case XK_Shift_L:	key = 0x10; break;
	case XK_Control_R:
	case XK_Control_L:	key = 0x11; break;
	case XK_Caps_Lock:	key = 0x14; break;
	case XK_Meta_L:		key = 0x5b; break;	/* windows key :) */
	case XK_Num_Lock:	key = 0x90; break;
	case XK_Pause:		key = 0xdf; break;	/* Suspend! */

	case XK_space:		key = 0x20; break;
	case XK_grave:		key = 0xc0; break;
	case XK_minus:		key = 0xbd; break;
	case XK_equal:		key = 0xbb; break;
	case XK_semicolon:	key = 0xba; break;
	case XK_comma:		key = 0xbc; break;
	case XK_period:		key = 0xbe; break;
	case XK_slash:		key = 0xbf; break;

	case XK_backslash:	key = 0xdc; break;
	case XK_apostrophe:	key = 0xde; break;
	case XK_bracketleft:	key = 0xdb; break;
	case XK_bracketright:	key = 0xdd; break;

	}

	if (key != 0x00) {
	    keyboard.status = htole16(event->type == KeyPress ?
					EVT_KEY_DOWN : EVT_KEY_UP);
	    keyboard.key = htole16(key);
	    writebufto(tw->cerdisp.clientsock, &keyboard, sizeof keyboard);
	} else {
	    if (event->type == KeyPress) {
		fprintf(stderr, "keypress <%s> not translated\n",
		    XKeysymToString(keysym));
		XBell(XtDisplay(w), 100);
	    }
	}
	break;
    default:
	fprintf(stderr, "InputAction: event %d not understood\n", event->type);
	return;
    }
}

/* Close the connection with the client (if open) */
static void
CloseAction(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    CerdispWidget tw = (CerdispWidget)w;
    int i;

    if (tw->cerdisp.connected) {
	close(tw->cerdisp.clientsock);
	tw->cerdisp.connected = 0;
	if (tw->cerdisp.clientinput != InvalidInputID) {
	    XtRemoveInput(tw->cerdisp.clientinput);
	    tw->cerdisp.clientinput = InvalidInputID;
	    if (tw->cerdisp.proxysock != -1) {
		close(tw->cerdisp.proxysock);
		XtRemoveInput(tw->cerdisp.proxyinput);
		tw->cerdisp.proxyinput = InvalidInputID;
		tw->cerdisp.proxysock = -1;
	    }
	}

	/* Release resources */
	if (tw->cerdisp.image != NULL) {
	    XDestroyImage(tw->cerdisp.image);
	    tw->cerdisp.image = NULL;
	}

	/* Free local memory resources if they had been allocated */
	XtFree((char *)tw->cerdisp.palette);
	XtFree(tw->cerdisp.buf);
	tw->cerdisp.palette = NULL;
	tw->cerdisp.buf = NULL;
	tw->cerdisp.bufsz = 0;

	/* Release colours of the old and new frames */
	lrufreeold(tw->cerdisp.lru);
	lrufreeold(tw->cerdisp.lru);

	clearwindow(tw);
	settitle(tw);

	/* Listen for further server connections again */
	for (i = 0; i < MAXSOCK; i++)
	    if (tw->cerdisp.serversock[i] != -1)
	        tw->cerdisp.serverinput[i] = XtAppAddInput(ac,
		    tw->cerdisp.serversock[i], (XtPointer)XtInputReadMask,
		    CerdispAccept, (XtPointer)tw);
    }
}

static void
ScaleAction(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    CerdispWidget tw = (CerdispWidget)w;
    char *inc;
    int incval;
    int newscale;
    Arg arglist[1];

    if (*num_params == 0)
	    inc = "+1";
    else
	    inc = params[0];
    incval = atoi(inc);

    /* skip leading space (XXX is this needed?) */
    while (inc[0] == ' ' || inc[0] == '\t')
	inc++;

    /* If the string starts with + or - then increment/decrement scale */
    if (inc[0] == '+' || inc[0] == '-')
	newscale = tw->cerdisp.scale + incval;
    else
	newscale = incval;

    XtSetArg(arglist[0], XtNscale, newscale);
    XtSetValues(w, arglist, 1);
}

/* Lookup a color in the palette */
static unsigned long
palette_lookup(CerdispWidget tw, unsigned int index)
{
    if (index >= (1 << tw->cerdisp.depth)) {
	XtAppWarning(ac, "palette_lookup(): index out of range");
	return 0;
    }
    return lrupixel(tw->cerdisp.lru, tw->cerdisp.palette[index].red,
	tw->cerdisp.palette[index].blue, tw->cerdisp.palette[index].green);
}
